import torch
import torch.nn as nn
import numpy as np
import os
from utils import gram_matrix

os.environ["KMP_DUPLICATE_LIB_OK"] = "TRUE"

class LosstextureModel(nn.Module):
    def gram_matrix(self,tensor):
        """ Calculate the Gram Matrix of a given tensor
        Gram Matrix: https://en.wikipedia.org/wiki/Gramian_matrix
        """
        # get the batch_size, depth, height, and width of the Tensor
        _, d, h, w = tensor.size()
        # reshape so we're multiplying the features for each channel
        Tensor = tensor.view(d, h * w)
        # calculate the gram matrix
        gram = torch.mm(Tensor, tensor.t())
        return gram


    def Ltexture(self, x,y):
        _, d, h, w = x.shape
        Ltext = 0
        for i in range(d):
            label_channel = x[:, i:i + 1]
            image_channel = y[:, i:i + 1]
            label_gram = gram_matrix(label_channel)
            image_gram = gram_matrix(image_channel)
            texture_loss = torch.mean((label_gram - image_gram) ** 2)
            Ltext += texture_loss
        Texture_Loss = Ltext / (d * h * w)
        return Texture_Loss
